import os

import flask
import yaml
from pywps import Service

from datacube.utils import import_function
from datacube.virtual import construct

from .startup_utils import setup_logger, initialise_prometheus, initialise_prometheus_register, setup_sentry


def create_process(process, input, **settings):
    process_class = import_function(process)
    return process_class(input=construct(**input), **settings)


def read_process_catalog(catalog_filename):
    with open(catalog_filename) as fl:
        config = yaml.load(fl, Loader=yaml.CLoader)

    return [create_process(**settings) for settings in config['processes']]


def create_app():
    # pylint: disable=unused-variable

    setup_logger()
    setup_sentry()

    app = flask.Flask(__name__)   # pylint: disable=redefined-outer-name
    app.url_map.strict_slashes = False

    metrics = initialise_prometheus(app)

    service = []

    @app.after_request
    def apply_cors(response):
        response.headers['Access-Control-Allow-Origin'] = '*'
        response.headers['Cache-Control'] = 'max-age=0'
        return response

    @app.route('/', methods=['GET', 'POST', 'HEAD'])
    def wps():
        if flask.request.method == 'HEAD':
            return ""

        if not service:
            service.append(Service(read_process_catalog('datacube-wps-config.yaml'), ['pywps.cfg']))
        return service[0]

    @app.route('/ping')
    def ping():
        return 'system is healthy'

    @app.route('/outputs/' + '<path:filename>')
    def outputfile(filename):
        targetfile = os.path.join('outputs', filename)
        if os.path.isfile(targetfile):
            file_ext = os.path.splitext(targetfile)[1]
            with open(targetfile, mode='rb') as f:
                file_bytes = f.read()
            mime_type = None
            if 'xml' in file_ext:
                mime_type = 'text/xml'
            return flask.Response(file_bytes, content_type=mime_type)

        flask.abort(404)

    # Note: register your default metrics after all routes have been set up.
    # Also note, that Gauge metrics registered as default will track the /metrics endpoint,
    # and this can't be disabled at the moment.
    initialise_prometheus_register(metrics)

    return app



from moto import mock_s3
import boto3
import altair as alt
from vega_datasets import data
import pywps.configuration as config

from datacube_wps.processes import upload_chart_svg_to_S3, upload_chart_html_to_S3

TEST_CHART = chart = (
        alt.Chart(data.cars.url)
        .mark_point()
        .encode(x="Horsepower:Q", y="Miles_per_Gallon:Q", color="Origin:N")
    )

TEST_CFG = "pywps.cfg"

@mock_s3
def test_s3_svg_chart_upload():
    config.load_configuration(TEST_CFG)
    bucket = config.get_config_value("s3", "bucket")
    region = config.get_config_value("s3", "region")
    location = {'LocationConstraint': region}
    client = boto3.client("s3",region_name=region)
    client.create_bucket(Bucket=bucket,CreateBucketConfiguration=location)
    upload_chart_svg_to_S3(TEST_CHART, "abcd")

@mock_s3
def test_s3_html_chart_upload():
    config.load_configuration(TEST_CFG)
    bucket = config.get_config_value("s3", "bucket")
    region = config.get_config_value("s3", "region")
    location = {'LocationConstraint': region}
    client = boto3.client("s3",region_name=region)
    client.create_bucket(Bucket=bucket,CreateBucketConfiguration=location)
    upload_chart_html_to_S3(TEST_CHART, "abcd")



import io

import flask
import pytest
from datacube.utils.geometry import Geometry

from datacube_wps.processes import _parse_geom, _get_time

TEST_FEATURE = {
    "features": [{
        "type": "Feature",
        "geometry": {"type": "Point", "coordinates": [125.6, 10.1]},
        "properties": {"name": "Dinagat Islands"},
        "crs": {"properties": {"name": "EPSG:4326"}},
    }]
}

class requests_mock_time:
    def __init__(self):
        self.inputs = {"start": [mock_date_data()], "end": [mock_date_data()]}

class mock_date_data:
    def __init__(self):
        self.data = """{
            "properties": {
                "timestamp" :
                {
                    "date-time" : "2021-02-01"
                }
            }
        }"""

def test_geom_parse():
    assert isinstance(_parse_geom(TEST_FEATURE), Geometry)

def test_time_parse():
    assert _get_time(requests_mock_time())



from datacube.virtual.transformations import ApplyMask
import numpy as np
import xarray as xr
import pandas as pd
from datacube.model import Measurement
from datacube.virtual.impl import Transformation
from . import geometry_mask, PolygonDrill
from datetime import datetime, timezone

ls_timezone = timezone.utc
def ls8_on(dataset):
    LS8_START_DATE = datetime(2013, 1, 1, tzinfo=ls_timezone)
    return dataset.center_time >= LS8_START_DATE

def ls7_on(dataset):
    LS7_STOP_DATE = datetime(2003, 5, 31, tzinfo=ls_timezone)
    LS7_STOP_AGAIN = datetime(2013, 5, 31, tzinfo=ls_timezone)
    LS7_START_AGAIN = datetime(2010, 1, 1, tzinfo=ls_timezone)
    return dataset.center_time <= LS7_STOP_DATE or (dataset.center_time >= LS7_START_AGAIN
                                                    and dataset.center_time <= LS7_STOP_AGAIN)

def ls5_on_1ym(dataset):
    LS5_START_AGAIN = datetime(2003, 1, 1, tzinfo=ls_timezone)
    LS5_STOP_DATE = datetime(1999, 12, 31, tzinfo=ls_timezone)
    LS5_STOP_AGAIN = datetime(2011, 12, 31, tzinfo=ls_timezone)
    return dataset.center_time <= LS5_STOP_DATE or (dataset.center_time >= LS5_START_AGAIN
                                                    and dataset.center_time <= LS5_STOP_AGAIN)

class WIT(PolygonDrill):
    def __init__(self, about, input, style):
        super().__init__(about, input, style)
        self.mask_all_touched = True
        print("mask all touch", self.mask_all_touched)
        
    def process_data(self, data, parameters):
        feature = parameters.get('feature')
        adays = parameters.get('aggregate', 0)
        geomask = geometry_mask(feature, data.geobox, invert=True, all_touched=self.mask_all_touched)
        #masked = mask_data(data)
        if adays > 0:
            aggregated = aggregate_over_time(data, adays)
        else:
            aggregated = data
        total_area = geomask.astype('int').sum()
        print("polygon area", total_area)
        re_wit = cal_area(aggregated) 
        re_wit = re_wit[(re_wit['valid']/total_area) > 0.9].dropna()
        re_wit = re_wit.drop(columns=['valid']).div(re_wit['valid'], axis=0)
        return re_wit
    
    def render_chart(self, df):
        pass

class TWnMask(Transformation):
    '''
    '''
    def __init__(self, category='wetness', coeffs=None):
        self.category = category
        if coeffs is None:
            self.coeffs = {
                 'brightness': {'blue': 0.2043, 'green': 0.4158, 'red': 0.5524, 'nir': 0.5741,
                                'swir1': 0.3124, 'swir2': 0.2303},
                 'greenness': {'blue': -0.1603, 'green': -0.2819, 'red': -0.4934, 'nir': 0.7940,
                               'swir1': -0.0002, 'swir2': -0.1446},
                 'wetness': {'blue': 0.0315, 'green': 0.2021, 'red': 0.3102, 'nir': 0.1594,
                             'swir1': -0.6806, 'swir2': -0.6109}
            }
        else:
            self.coeffs = coeffs
        self.var_name = f'TC{category[0].upper()}'

    def compute(self, data):
        tci_var = 0
        for key in self.coeffs[self.category].keys():
            nodata = getattr(data[key], 'nodata', -1)
            data[key] = data[key].where(data[key] > nodata)
            tci_var += data[key] * self.coeffs[self.category][key]
        tci_var.data[np.isnan(tci_var.data)] = -9999
        tci_var = tci_var.astype(np.float32)
        tci_var.attrs = dict(nodata=-9999, units=1, crs=data.attrs['crs'])
        tci_var = tci_var.to_dataset(name=self.var_name)
        
        def make_mask():
            water_value = (((data.water & (1 << 7)) != 0)
                           & (((data.water & (1 << 6)) | (data.water & (1 << 5)) | (data.water & (1 << 1))
                               | (data.water & 1)) == 0)).astype('int16')
            water_value.attrs = dict(nodata=0, units=1, crs=data.attrs['crs'])
            water_value = water_value.to_dataset(name='water')

            pmask = (((data.fmask != 2) & (data.fmask != 3) & (data.fmask != 0) & (data.nbart_contiguity == 1))
                     & (((data.water & (1 << 7)) | (data.water & (1 << 6)) | (data.water & (1 << 5))
                         | (data.water & (1 << 1)) | (data.water & 1)) == 0))
            pmask.attrs = dict(nodata=False, units=1, crs=data.attrs['crs'])
            pmask = pmask.to_dataset(name='pmask')
            return pmask.merge(water_value)
        pmask = tci_var.merge(make_mask())
        pmask.attrs = data.attrs
        return pmask

    def measurements(self, input_measurements):
        return {self.var_name: Measurement(name=self.var_name, dtype='float32', nodata=-9999, units='1'),
               'water': Measurement(name='water', dtype='int16', nodata=0, units='1'),
               'pmask': Measurement(name='pmask', dtype='bool', nodata=False, units='1')}
    
def mask_data(loaded):
    def make_mask():
        # water mask
        # flags = {'cloud': False,
        #    'cloud_shadow': False,
        #    'noncontiguous': False,
        #    'water_observed': False,
        #    'nodata': False
        #    }                                                                                                                                 
        # water value
        # flags = {'cloud': False,
        #    'cloud_shadow': False,
        #    'noncontiguous': False,
        #    'water_observed': True,
        #    'nodata': False
        #    }
        water_value = (((loaded.water & (1 << 7)) != 0)
                       & (((loaded.water & (1 << 6)) | (loaded.water & (1 << 5)) | (loaded.water & (1 << 1))
                           | (loaded.water & 1)) == 0)).astype('int16').to_dataset(name='water')
        pmask = (((loaded.fmask != 2) & (loaded.fmask != 3) & (loaded.fmask != 0) & (loaded.nbart_contiguity == 1))
                 & (((loaded.water & (1 << 7)) | (loaded.water & (1 << 6)) | (loaded.water & (1 << 5))
                     | (loaded.water & (1 << 1)) | (loaded.water & 1)) == 0)).to_dataset(name='pmask')
        return pmask.merge(water_value)
        
    pmask = make_mask()
    loaded = loaded.drop(['fmask', 'nbart_contiguity', 'water']).merge(pmask)
    loaded = ApplyMask('pmask', apply_to=['bs', 'pv', 'npv', 'TCW']).compute(loaded)
    return loaded

def average_over_day(masked):
    tmp = xr.Dataset()
    for var in masked.data_vars:
        if var != 'water':
            valid_var = ~(np.abs(masked[var] - masked[var].attrs['nodata']) < 1e-5)
            # to suppress the warnings from dask if using `mean`
            # haven't got another way
            tmp[var] = (masked[var].where(valid_var).sum('time', skipna=False, keep_attrs=True) / masked.time.shape[0]).fillna(masked[var].attrs['nodata']).astype('float32')
        else:
            tmp[var] = masked[var].astype('int16').sum('time', keep_attrs=True).astype('int16')
        tmp[var].attrs = masked[var].attrs
    tmp = tmp.expand_dims(time=masked.time[0:1], axis=0)
    tmp.attrs = masked.attrs
    return tmp

def aggregate_data(masked):    
    tmp = masked[dict(time=[0])].copy(deep=True)
    for time in masked.time.data[1:]:
        for var in masked.data_vars:
            if var != 'water':
                valid_var = ~(np.abs(tmp[var] - masked[var].attrs['nodata']) < 1e-5) | tmp.water > 0
            else:
                valid_var = ~(np.abs(tmp.TCW - masked.TCW.attrs['nodata']) < 1e-5) | tmp.water > 0
            tmp[var] = tmp[var].where(valid_var, 0) + masked[var].sel(time=time).where(~valid_var, 0)
            tmp[var].attrs = masked[var].attrs
    return tmp

def aggregate_over_time(masked, days):
    i_start = 0
    i_end = i_start + 1
    aggregated = None
    print("aggregate over days", days)
    while i_end < masked.time.size:
        time = masked.time.data[i_start]
        while (np.abs(time - masked.time.data[i_end]).astype('timedelta64[D]')
                                                 < np.timedelta64(days, 'D')):
            i_end += 1
            if i_end >= masked.time.size:
                break
        if days > 1:
            tmp = aggregate_data(masked[dict(time=np.arange(i_start, i_end))])
        else:
            tmp = average_over_day(masked[dict(time=np.arange(i_start, i_end))])
        if aggregated is None:
            aggregated = tmp
        else:
            aggregated = xr.concat([aggregated, tmp], dim='time')
        i_start = i_end
        i_end = i_start + 1
    print("finish aggregating", datetime.now())
    return aggregated

def cal_area(aggregated, wet_threshold=-350):
    non_columns = ['spatial_ref']
    re = aggregated.water.sum(dim=['x', 'y']).load().to_dataframe(name='water').drop(columns=non_columns)
    valid = np.abs(aggregated.TCW - aggregated.TCW.attrs['nodata']) > 1e-5
    tcw = valid.astype('int').sum(dim=['x', 'y']).load().to_dataframe(name='tcw').drop(columns=non_columns)
    re.insert(0, 'valid', re['water'] + tcw['tcw'])
    re = pd.merge(re,(aggregated.TCW > wet_threshold).astype('int').sum(dim=['x', 'y']).load().to_dataframe(name='wet').drop(columns=non_columns),
                  on=['time'], how='inner')
    fc_com = (aggregated[['bs', 'pv', 'npv']].where(((aggregated.TCW < wet_threshold) & valid), 0) / 100).sum(dim=['x', 'y']).load()
    for var in fc_com.data_vars:
        re = pd.merge(re, fc_com[var].to_dataframe(name=var).drop(columns=non_columns), on=['time'], how='inner')
    return re


import pytest

from datacube.utils.geometry import CRS, Geometry

from datacube_wps.impl import read_process_catalog
from datacube_wps.processes.fcdrill import FCDrill
from datacube_wps.processes.mangrovedrill import MangroveDrill
from datacube_wps.processes.wofsdrill import WOfSDrill


def test_fc():
    catalog = read_process_catalog("datacube-wps-config.yaml")
    fc = [entry for entry in catalog if isinstance(entry, FCDrill)][0]
    poly = Geometry(
        {
            "type": "Polygon",
            "coordinates": [
                [
                    (147.28271484375003, -35.89238773935897),
                    (147.03277587890628, -35.663990911348115),
                    (146.65237426757815, -35.90684930677119),
                    (147.09182739257815, -36.15894422111004),
                    (147.28271484375003, -35.89238773935897),
                ]
            ],
        },
        crs=CRS("EPSG:4326"),
    )
    results = fc.query_handler(time=("2019-03-05", "2019-07-10"), feature=poly)
    assert "data" in results
    assert "chart" in results


def test_mangrove():
    catalog = read_process_catalog("datacube-wps-config.yaml")
    fc = [entry for entry in catalog if isinstance(entry, MangroveDrill)][0]
    poly = Geometry(
        {
            "type": "Polygon",
            "coordinates": [
                [
                    (143.98956298828125, -14.689881366618762),
                    (144.26422119140625, -14.689881366618762),
                    (144.26422119140625, -14.394778454856146),
                    (143.98956298828125, -14.394778454856146),
                    (143.98956298828125, -14.689881366618762),
                ]
            ],
        },
        crs=CRS("EPSG:4326"),
    )
    results = fc.query_handler(time=("2000", "2005"), feature=poly)
    assert "data" in results
    assert "chart" in results

#@pytest.mark.xfail(reason="Pixel drills need special config")
def test_wofs():
    catalog = read_process_catalog("datacube-wps-config.yaml")
    wofs = [entry for entry in catalog if isinstance(entry, WOfSDrill)][0]
    point = Geometry(
        {"type": "Point", "coordinates": [137.01475095074406, -28.752777955850917, 0]},
        crs=CRS("EPSG:4326"),
    )
    results = wofs.query_handler(time="2000", feature=point)
    assert "data" in results
    assert "chart" in results



import boto3
from moto import mock_s3

from datacube_wps.impl import create_app

def test_ping(client):
    r = client.get('/ping')
    assert r.status_code == 200


def test_head(client):
    r = client.head('/')
    assert r.status_code == 200


@mock_s3
def xtest_mangrove(client):
    conn = boto3.resource('s3', region_name='ap-southeast-2')
    conn.create_bucket(Bucket='dea-wps-results', CreateBucketConfiguration={'LocationConstraint': 'ap-southeast-2'})

    headers = {'Content-Type': 'text/xml;charset=UTF-8', 'cache-control': 'max-age=0'}

    data = """<?xml version="1.0" encoding="UTF-8"?>
    <wps:Execute version="1.0.0" service="WPS" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns="http://www.opengis.net/wps/1.0.0" xmlns:wfs="http://www.opengis.net/wfs" xmlns:wps="http://www.opengis.net/wps/1.0.0" xmlns:ows="http://www.opengis.net/ows/1.1" xmlns:gml="http://www.opengis.net/gml" xmlns:ogc="http://www.opengis.net/ogc" xmlns:wcs="http://www.opengis.net/wcs/1.1.1" xmlns:xlink="http://www.w3.org/1999/xlink" xsi:schemaLocation="http://www.opengis.net/wps/1.0.0 http://schemas.opengis.net/wps/1.0.0/wpsAll.xsd">
      <ows:Identifier>WOfSDrill</ows:Identifier>
      <wps:DataInputs>
        <wps:Input>
          <ows:Identifier>geometry</ows:Identifier>
          <wps:Data>
            <wps:ComplexData>{"type":"FeatureCollection","features":[{"type":"Feature","geometry":{"type":"Point","coordinates":[146.85029736971987,-32.94459759906837,-1822.7196235501208]}}]}</wps:ComplexData>
          </wps:Data>
        </wps:Input>
      </wps:DataInputs>
      <wps:ResponseForm>
        <wps:ResponseDocument storeExecuteResponse="true" status="true"/>
      </wps:ResponseForm>
    </wps:Execute>
    """

    r = client.post('/?service=WPS&request=Execute', headers=headers, data=data)
    assert r.status_code == 200



import numpy as np
import datacube
import altair

from pywps import LiteralOutput, ComplexInput, ComplexOutput

from . import PixelDrill, FORMATS, log_call, chart_dimensions


class WOfSDrill(PixelDrill):
    def input_formats(self):
        return [ComplexInput('geometry', 'Location (Lon, Lat)', supported_formats=[FORMATS['point']])]

    def output_formats(self):
        return [LiteralOutput("image", "WOfS Pixel Drill Preview"),
                LiteralOutput("url", "WOfS Pixel Drill Graph"),
                ComplexOutput('timeseries', 'Timeseries Drill', supported_formats=[FORMATS['output_json']])]

    @log_call
    def process_data(self, data, parameters):
        # TODO raise ProcessError('query returned no data') when appropriate

        rules = [
            {
                'op': any,
                'flags': ['terrain_or_low_angle', 'cloud_shadow', 'cloud', 'high_slope', 'noncontiguous'],
                'value': 'not observable'
            },
            {
                'op': all,
                'flags': ['dry', 'sea'],
                'value': 'not observable'
            },
            {
                'op': any,
                'flags': ['dry'],
                'value': 'dry'
            },
            {
                'op': any,
                'flags': ['wet', 'sea'],
                'value': 'wet'
            }
        ]

        # TODO: investigate why PixelDrill is changing datatype
        water = data.data_vars['water']
        data['observation'] = water.astype('int16')
        data = data.drop_vars(['water'])

        def get_flags(val):
            flag_dict = datacube.utils.masking.mask_to_dict(water.attrs['flags_definition'], val)
            flags = list(filter(flag_dict.get, flag_dict))
            # apply rules in sequence
            ret_val = 'not observable'
            for rule in rules:
                if rule['op']([r in flags for r in rule['flags']]):
                    ret_val = rule['value']
                    break
            return ret_val
        gf = np.vectorize(get_flags)

        data['observation'].values = gf(data['observation'].values)

        df = data.to_dataframe()
        df.reset_index(inplace=True)
        return df

    @log_call
    def render_chart(self, df):
        width, height = chart_dimensions(self.style)

        pt_lat = df['latitude'].iat[0]
        pt_lon = df['longitude'].iat[0]

        yscale = altair.Scale(domain=['wet', 'dry', 'not observable'])
        ascale = altair.Scale(domain=['wet', 'dry', 'not observable'],
                              range=['blue', 'red', 'grey'])
        chart = altair.Chart(df,
                             width=width,
                             height=height,
                             autosize='fit',
                             background='white',
                             title=f"Water Observations for {pt_lat:.6f},{pt_lon:.6f}")
        chart = chart.mark_tick(thickness=3)
        chart = chart.encode(x=altair.X('time:T'),
                             y=altair.Y('observation:nominal', scale=yscale),
                             color=altair.Color('observation:nominal', scale=ascale),
                             tooltip=['observation:nominal',
                                      altair.Tooltip(field='time',
                                                     format='%d %B, %Y',
                                                     title='Date',
                                                     type='temporal')])
        return chart

    @log_call
    def render_outputs(self, df, chart):
        return super().render_outputs(df, chart,
                                      is_enabled=False, name="WOfS", header=['Observation'])



#!/usr/bin/env python3

import os
import argparse

from .impl import create_app


app = create_app()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Script for starting an datacube-wps instance")
    parser.add_argument('-d', '--daemon',
                        action='store_true', help="run in daemon mode")
    parser.add_argument('-a', '--all-addresses',
                        action='store_true', help=("run flask using IPv4 0.0.0.0 (all network interfaces), "
                                                   "otherwise bind to 127.0.0.1 (localhost). "
                                                   "This maybe necessary in systems that only run Flask"))
    args = parser.parse_args()

    if args.all_addresses:
        bind_host = '0.0.0.0'
    else:
        bind_host = '127.0.0.1'

    if args.daemon:
        pid = None
        try:
            pid = os.fork()
        except OSError as e:
            raise Exception("%s [%d]" % (e.strerror, e.errno)) from e

        if pid == 0:
            os.setsid()
            app.run(threaded=True, host=bind_host)
        else:
            os._exit(0)  # pylint: disable=protected-access
    else:
        app.run(threaded=True, host=bind_host)



import os
import logging

from prometheus_flask_exporter.multiprocess import GunicornInternalPrometheusMetrics
import sentry_sdk
from sentry_sdk.integrations.flask import FlaskIntegration
from flask import request

LOG_FORMAT = ('%(asctime)s] [%(levelname)s] file=%(pathname)s line=%(lineno)s '
              'module=%(module)s function=%(funcName)s %(message)s')


def setup_logger():
    logger = logging.getLogger('PYWPS')
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(LOG_FORMAT))
    logger.addHandler(handler)


def initialise_prometheus(app, log=None):
    if os.environ.get("prometheus_multiproc_dir", False):
        metrics = GunicornInternalPrometheusMetrics(app)
        if log:
            log.info("Prometheus metrics enabled")
        return metrics
    return None


def initialise_prometheus_register(metrics):
    # Register routes with Prometheus - call after all routes set up.
    if os.environ.get("prometheus_multiproc_dir", False):
        metrics.register_default(
            metrics.summary(
                'flask_wps_request_full_url', 'Request summary by request url',
                labels={
                    'query_request': lambda: request.args.get('request'),
                    'query_url': lambda: request.full_path
                }
            )
        )


def setup_sentry():
    if os.environ.get("SENTRY_KEY") and os.environ.get("SENTRY_PROJECT"):
        SENTRY_ENV_TAG = os.environ.get("SENTRY_ENV_TAG") if os.environ.get("SENTRY_ENV_TAG") else "dev"
        sentry_sdk.init(
            dsn="https://%s@sentry.io/%s" % (os.environ["SENTRY_KEY"], os.environ["SENTRY_PROJECT"]),
            environment=SENTRY_ENV_TAG,
            integrations = [FlaskIntegration()]
        )


import xarray
import altair
from . import PolygonDrill, log_call, chart_dimensions


class MangroveDrill(PolygonDrill):

    @log_call
    def process_data(self, data, parameters):
        data = data.compute()

        # TODO raise ProcessError('query returned no data') when appropriate
        woodland = data.where(data == 1).count(['x', 'y'])
        woodland = woodland.rename(name_dict={'canopy_cover_class': 'Woodland'})
        open_forest = data.where(data == 2).count(['x', 'y'])
        open_forest = open_forest.rename(name_dict={'canopy_cover_class': 'Open Forest'})
        closed_forest = data.where(data == 3).count(['x', 'y'])
        closed_forest = closed_forest.rename(name_dict={"canopy_cover_class": 'Closed Forest'})

        final = xarray.merge([woodland, open_forest, closed_forest])
        result = final.to_dataframe()
        result = result.drop('spatial_ref', axis=1)
        result.reset_index(inplace=True)
        return result

    @log_call
    def render_chart(self, df):
        width, height = chart_dimensions(self.style)

        melted = df.melt('time', var_name='Cover Type', value_name='Area')
        melted = melted.dropna()

        style = self.style['table']['columns']
        cover_types = ['Woodland', 'Open Forest', 'Closed Forest']

        chart = altair.Chart(melted,
                             width=width,
                             height=height,
                             title='Percentage of Area - Mangrove Canopy Cover')
        chart = chart.mark_area()
        chart = chart.encode(x='time:T',
                             y=altair.Y('Area:Q', stack='normalize'),
                             color=altair.Color('Cover Type:N',
                                                scale=altair.Scale(domain=cover_types,
                                                                   range=[style[ct]['chartLineColor']
                                                                          for ct in cover_types])),
                             tooltip=[altair.Tooltip(field='time', format='%d %B, %Y', title='Date', type='temporal'),
                                      'Area:Q',
                                      'Cover Type:N'])

        return chart

    @log_call
    def render_outputs(self, df, chart):
        return super().render_outputs(df, chart, is_enabled=True, name="Mangrove Cover",
                                      header=['Woodland', 'Open Forest', 'Closed Forest'])



from functools import wraps
from timeit import default_timer
import json
import io
import os

from pywps import Process, ComplexInput, ComplexOutput, Format
from pywps.app.exceptions import ProcessError
import pywps.configuration as config

import datacube
from datacube.utils.geometry import Geometry, CRS
from datacube.api.core import output_geobox, query_group_by

from datacube.drivers import new_datasource
from datacube.storage import BandInfo
from datacube.utils.rio import configure_s3_access

import rasterio.features

import xarray
import numpy as np
import pandas
import altair
from dask.distributed import Client

import boto3
import botocore
from botocore.client import Config

from dateutil.parser import parse


FORMATS = {
    # Defines the format for the returned object
    # in this case a JSON object containing a CSV
    'output_json': Format('application/vnd.terriajs.catalog-member+json',
                          schema='https://tools.ietf.org/html/rfc7159',
                          encoding="utf-8"),
    'point': Format('application/vnd.geo+json',
                    schema='http://geojson.org/geojson-spec.html#point'),
    'polygon': Format('application/vnd.geo+json',
                      schema='http://geojson.org/geojson-spec.html#polygon'),
    'datetime': Format('application/vnd.geo+json',
                       schema='http://www.w3.org/TR/xmlschema-2/#dateTime')
}

GB = 1.e9
MAX_BYTES_IN_GB = 20.0
MAX_BYTES_PER_OBS_IN_GB = 2.0


def log_call(func):
    @wraps(func)
    def log_wrapper(*args, **kwargs):
        name = func.__name__
        for index, arg in enumerate(args):
            try:
                arg_name = func.__code__.co_varnames[index]
            except (AttributeError, KeyError, IndexError):
                arg_name = f'arg #{index}'
            print(f'{name} {arg_name}: {arg}')
        for key, value in kwargs.items():
            print(f'{name} {key}: {value}')

        start = default_timer()
        result = func(*args, **kwargs)
        end = default_timer()
        print('{} returned {}'.format(name, result))
        print('{} took {:.3f}s'.format(name, end - start))
        return result

    return log_wrapper


@log_call
def _uploadToS3(filename, data, mimetype):
    session = boto3.Session()
    bucket = config.get_config_value('s3', 'bucket')
    s3 = session.client('s3')
    s3.upload_fileobj(data,
                      bucket,
                      filename,
                      ExtraArgs={'ACL': 'public-read', 'ContentType': mimetype})

    # Create unsigned s3 client for determining public s3 url
    s3 = session.client('s3', config=Config(signature_version=botocore.UNSIGNED))
    return s3.generate_presigned_url(ClientMethod='get_object',
                                     ExpiresIn=0,
                                     Params={'Bucket': bucket, 'Key': filename})


def upload_chart_html_to_S3(chart : altair.Chart, process_id : str):
    html_io = io.StringIO()
    chart.save(html_io, format='html')
    html_bytes = io.BytesIO(html_io.getvalue().encode())
    return _uploadToS3(process_id + '/chart.html', html_bytes, 'text/html')


def upload_chart_svg_to_S3(chart : altair.Chart, process_id : str):
    img_io = io.StringIO()
    chart.save(img_io, format='svg')
    img_bytes = io.BytesIO(img_io.getvalue().encode())
    return _uploadToS3(process_id + '/chart.svg', img_bytes, 'image/svg+xml')


# From https://stackoverflow.com/a/16353080
class DatetimeEncoder(json.JSONEncoder):
    def default(self, obj):
        try:
            return super(DatetimeEncoder, obj).default(obj)
        except TypeError:
            return str(obj)


def geometry_mask(geom, geobox, all_touched=False, invert=False):
    return rasterio.features.geometry_mask([geom.to_crs(geobox.crs)],
                                           out_shape=geobox.shape,
                                           transform=geobox.affine,
                                           all_touched=all_touched,
                                           invert=invert)


def wofls_fuser(dest, src):
    where_nodata = (src & 1) == 0
    np.copyto(dest, src, where=where_nodata)
    return dest


def chart_dimensions(style):
    if 'chart' in style and 'width' in style['chart']:
        width = style['chart']['width']
    else:
        width = 1000
    if 'height' in style and 'height' in style['chart']:
        height = style['chart']['height']
    else:
        height = 300

    return (width, height)


def _guard_rail(input, box):
    measurement_dicts = input.output_measurements(box.product_definitions)

    byte_count = 1
    for x in box.shape:
        byte_count *= x
    byte_count *= sum(np.dtype(m.dtype).itemsize for m in measurement_dicts.values())

    print('byte count for query: ', byte_count)
    if byte_count > MAX_BYTES_IN_GB * GB:
        raise ProcessError(("requested area requires {}GB data to load - "
                            "maximum is {}GB").format(int(byte_count / GB), MAX_BYTES_IN_GB))

    try:
        grouped = box.box
    except AttributeError:
        # datacube 1.7 compatibility
        grouped = box.pile

    print('grouped shape', grouped.shape)
    assert len(grouped.shape) == 1

    if grouped.shape[0] == 0:
        raise ProcessError('no data returned for query')

    bytes_per_obs = byte_count / grouped.shape[0]
    if bytes_per_obs > MAX_BYTES_PER_OBS_IN_GB * GB:
        raise ProcessError(("requested time slices each requires {}GB data to load - "
                            "maximum is {}GB").format(int(bytes_per_obs / GB), MAX_BYTES_PER_OBS_IN_GB))


def _datetimeExtractor(data):
    return parse(json.loads(data)["properties"]["timestamp"]["date-time"])

def _parse_geom(request_json):
    features = request_json['features']
    if len(features) < 1:
        # can't drill if there is no geometry
        raise ProcessError("no features specified")

    if len(features) > 1:
        # do we need multipolygon support here?
        raise ProcessError("multiple features specified")

    feature = features[0]

    if hasattr(request_json, 'crs'):
        crs = CRS(request_json['crs']['properties']['name'])
    elif hasattr(feature, 'crs'):
        crs = CRS(feature['crs']['properties']['name'])
    else:
        # http://geojson.org/geojson-spec.html#coordinate-reference-system-objects
        crs = CRS('urn:ogc:def:crs:OGC:1.3:CRS84')

    return Geometry(feature['geometry'], crs)

def _get_feature(request):
    stream = request.inputs['geometry'][0].stream
    request_json = json.loads(stream.readline())

    return _parse_geom(request_json)


def _get_time(request):
    if 'start' not in request.inputs or 'end' not in request.inputs:
        return None

    def _datetimeExtractor(data):
        return parse(json.loads(data)["properties"]["timestamp"]["date-time"])

    return (_datetimeExtractor(request.inputs['start'][0].data),
            _datetimeExtractor(request.inputs['end'][0].data))


def _get_parameters(request):
    if 'parameters' not in request.inputs:
        return {}

    stream = request.inputs['parameters'][0].stream
    params = json.loads(stream.readline())

    return params


def _render_outputs(uuid, style, df: pandas.DataFrame, chart: altair.Chart,
                    is_enabled=True, name="Timeseries", header=True):
    html_url = upload_chart_html_to_S3(chart, str(uuid))
    img_url = upload_chart_svg_to_S3(chart, str(uuid))

    try:
        csv_df = df.drop(columns=['latitude', 'longitude'])
    except KeyError:
        csv_df = df

    csv_df.set_index('time', inplace=True)
    csv = csv_df.to_csv(header=header, date_format="%Y-%m-%d")

    if 'table' in style:
        table_style = {'tableStyle': style['table']}
    else:
        table_style = {}

    output_dict = {
        "data": csv,
        "isEnabled": is_enabled,
        "type": "csv",
        "name": name,
        **table_style
    }

    output_json = json.dumps(output_dict, cls=DatetimeEncoder)

    outputs = {
        'image': {'data': img_url},
        'url': {'data': html_url},
        'timeseries': {'data': output_json}
    }

    return outputs


def _populate_response(response, outputs):
    for ident, output_value in outputs.items():
        if ident in response.outputs:
            if "data" in output_value:
                response.outputs[ident].data = output_value['data']
            if "output_format" in output_value:
                response.outputs[ident].output_format = output_value['output_format']
            if "url" in output_value:
                response.outputs[ident].url = output_value['url']


def num_workers():
    return int(os.getenv('DATACUBE_WPS_NUM_WORKERS', '4'))


class PixelDrill(Process):
    def __init__(self, about, input, style):
        if 'geometry_type' in about:
            assert about['geometry_type'] == 'point'

        super().__init__(handler=self.request_handler,
                         inputs=self.input_formats(),
                         outputs=self.output_formats(),
                         **{key: value for key, value in about.items() if key not in ['geometry_type']})

        self.about = about
        self.input = input
        self.style = style

    def input_formats(self):
        return [ComplexInput('geometry', 'Location (Lon, Lat)', supported_formats=[FORMATS['point']]),
                ComplexInput('start', 'Start Date', supported_formats=[FORMATS['datetime']]),
                ComplexInput('end', 'End date', supported_formats=[FORMATS['datetime']])]

    def output_formats(self):
        return [ComplexOutput('timeseries', 'Timeseries Drill',
                              supported_formats=[FORMATS['output_json']])]

    def request_handler(self, request, response):
        time = _get_time(request)
        feature = _get_feature(request)
        parameters = _get_parameters(request)

        result = self.query_handler(time, feature, parameters=parameters)

        outputs = self.render_outputs(result['data'], result['chart'])
        _populate_response(response, outputs)
        return response

    @log_call
    def query_handler(self, time, feature, parameters=None):
        if parameters is None:
            parameters = {}

        with Client(n_workers=1, processes=False, threads_per_worker=num_workers()) as client:
            configure_s3_access(aws_unsigned=True,
                                region_name=os.getenv('AWS_DEFAULT_REGION', 'auto'),
                                client=client)

            with datacube.Datacube() as dc:
                data = self.input_data(dc, time, feature)

        df = self.process_data(data, {'time': time, 'feature': feature, **parameters})
        chart = self.render_chart(df)

        return {'data': df, 'chart': chart}

    @log_call
    def input_data(self, dc, time, feature):
        if time is None:
            bag = self.input.query(dc, geopolygon=feature)
        else:
            bag = self.input.query(dc, time=time, geopolygon=feature)

        lonlat = feature.coords[0]

        measurements = self.input.output_measurements(bag.product_definitions)
        box = self.input.group(bag)

        data = self.input.fetch(box, dask_chunks={'time': 1})
        data = data.compute()

        coords = {'longitude': np.array([lonlat[0]]),
                  'latitude': np.array([lonlat[1]]),
                  'time': data.time.data}

        result = xarray.Dataset()
        for measurement_name, measurement in measurements.items():
            result[measurement_name] = xarray.DataArray(data[measurement_name],
                                                        dims=('time', 'longitude', 'latitude'),
                                                        coords=coords,
                                                        attrs={key: value
                                                               for key, value in measurement.items()
                                                               if key in ['flags_definition']})
        return result

    def process_data(self, data: xarray.Dataset, parameters: dict) -> pandas.DataFrame:
        raise NotImplementedError

    def render_chart(self, df: pandas.DataFrame) -> altair.Chart:
        raise NotImplementedError

    def render_outputs(self, df: pandas.DataFrame, chart: altair.Chart,
                       is_enabled=True, name="Timeseries", header=True):
        return _render_outputs(self.uuid, self.style, df, chart,
                               is_enabled=is_enabled, name=name, header=header)


class PolygonDrill(Process):
    def __init__(self, about, input, style):
        if 'geometry_type' in about:
            assert about['geometry_type'] == 'polygon'

        super().__init__(handler=self.request_handler,
                         inputs=self.input_formats(),
                         outputs=self.output_formats(),
                         **{key: value for key, value in about.items() if key not in ['geometry_type']})

        self.about = about
        self.input = input
        self.style = style
        self.mask_all_touched = False

    def input_formats(self):
        return [ComplexInput('geometry', 'Geometry', supported_formats=[FORMATS['polygon']]),
                ComplexInput('start', 'Start Date', supported_formats=[FORMATS['datetime']]),
                ComplexInput('end', 'End date', supported_formats=[FORMATS['datetime']])]

    def output_formats(self):
        return [ComplexOutput('timeseries', 'Timeseries Drill',
                              supported_formats=[FORMATS['output_json']], as_reference=False)]

    def request_handler(self, request, response):
        time = _get_time(request)
        feature = _get_feature(request)
        parameters = _get_parameters(request)

        result = self.query_handler(time, feature, parameters=parameters)

        outputs = self.render_outputs(result['data'], result['chart'])
        _populate_response(response, outputs)
        return response

    @log_call
    def query_handler(self, time, feature, parameters=None):
        if parameters is None:
            parameters = {}

        with Client(n_workers=num_workers(), processes=True, threads_per_worker=1) as client:
            configure_s3_access(aws_unsigned=True,
                                region_name=os.getenv('AWS_DEFAULT_REGION', 'auto'),
                                client=client)

            with datacube.Datacube() as dc:
                data = self.input_data(dc, time, feature)

        df = self.process_data(data, {'time': time, 'feature': feature, **parameters})
        chart = self.render_chart(df)

        return {'data': df, 'chart': chart}

    def input_data(self, dc, time, feature):
        if time is None:
            bag = self.input.query(dc, geopolygon=feature)
        else:
            bag = self.input.query(dc, time=time, geopolygon=feature)

        box = self.input.group(bag)
        # _guard_rail(self.input, box)
        # TODO customize the number of processes
        data = self.input.fetch(box, dask_chunks={'time': 1})
        mask = geometry_mask(feature, data.geobox,
                             all_touched=self.mask_all_touched, invert=True)
        
        # mask out data outside requested polygon
        for band_name, band_array in data.data_vars.items():
            if 'nodata' in band_array.attrs:
                data[band_name] = band_array.where(mask, other=band_array.attrs['nodata'])
            else:
                data[band_name] = band_array.where(mask)

        return data

    def process_data(self, data: xarray.Dataset, parameters: dict) -> pandas.DataFrame:
        raise NotImplementedError

    def render_chart(self, df: pandas.DataFrame) -> altair.Chart:
        raise NotImplementedError

    def render_outputs(self, df: pandas.DataFrame, chart: altair.Chart,
                       is_enabled=True, name="Timeseries", header=True):
        return _render_outputs(self.uuid, self.style, df, chart,
                               is_enabled=is_enabled, name=name, header=header)



from setuptools import setup, find_packages

with open('VERSION.txt') as ff:
    VERSION = ff.read().strip()

INSTALL_REQUIRES = [
    'automat >= 0.3.0',
    'flask',
    'scikit-image',
    'gunicorn',
    'datacube',
    'rasterio >= 1.0.9',
    'rasterio[s3]',
    'altair',
    'altair_saver',
    'selenium',
    'python-dateutil',
    'sentry_sdk',
    'prometheus-flask-exporter',
    'blinker',
    'pywps == 4.2.4'
]

DESCRIPTION = ("datacube-wps is an implementation of the Web Processing Service standard "
               "from the Open Geospatial Consortium.")

KEYWORDS = 'ODC WPS OGC processing'

config = {
    'description': DESCRIPTION,
    'keywords': KEYWORDS,
    'author': 'Digital Earth Australia',
    'license': 'MIT',
    'platforms': 'all',
    'url': 'http://www.ga.gov.au/dea',
    'download_url': 'https://github.com/opendatacube/datacube-wps',
    'author_email': 'earth.observation@ga.gov.au',
    'maintainer': 'Digital Earth Australia',
    'maintainer_email': 'earth.observation@ga.gov.au',
    'classifiers': [
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: GIS'
    ],
    'version': VERSION,
    'install_requires': INSTALL_REQUIRES,
    'packages': find_packages(),
    'name': 'datacube-wps'
}

setup(**config)



from timeit import default_timer

import altair
import xarray
import numpy as np

from datacube.utils.masking import make_mask

from pywps import LiteralOutput, ComplexOutput

from . import PolygonDrill, log_call, chart_dimensions, FORMATS


class FCDrill(PolygonDrill):
    SHORT_NAMES = ['BS', 'PV', 'NPV', 'Unobservable']
    LONG_NAMES = ['Bare Soil',
                  'Photosynthetic Vegetation',
                  'Non-Photosynthetic Vegetation',
                  'Unobservable']

    def output_formats(self):
        return [LiteralOutput("image", "Fractional Cover Drill Preview"),
                LiteralOutput("url", "Fractional Cover Drill Chart"),
                ComplexOutput('timeseries', 'Fractional Cover Polygon Drill Timeseries', supported_formats=[FORMATS['output_json']])]

    @log_call
    def process_data(self, data, parameters):
        wofs_mask_flags = [
            dict(dry=True),
            dict(terrain_or_low_angle=False, high_slope=False, cloud_shadow=False, cloud=False, sea=False)
        ]

        water = data.data_vars['water']
        data = data.drop_vars(['water'])

        total = data.count(dim=['x', 'y'])
        total_valid = (data != -1).sum(dim=['x', 'y'])

        # TODO enable this check, investigate why it fails
        # if total_valid <= 0:
        #     raise ProcessError('query returned no data')

        for m in wofs_mask_flags:
            mask = make_mask(water, **m)
            data = data.where(mask)

        total_invalid = (np.isnan(data)).sum(dim=['x', 'y'])
        not_pixels = total_valid - (total - total_invalid)

        # following robbi's advice, cast the dataset to a dataarray
        maxFC = data.to_array(dim='variable', name='maxFC')

        # turn FC array into integer only as nanargmax doesn't seem to handle floats the way we want it to
        FC_int = maxFC.astype('int16')

        # use numpy.nanargmax to get the index of the maximum value along the variable dimension
        # BSPVNPV=np.nanargmax(FC_int, axis=0)
        BSPVNPV = FC_int.argmax(dim='variable')

        FC_mask = np.isfinite(maxFC).all(dim='variable')   # pylint: disable=no-member,unexpected-keyword-arg

        # #re-mask with nans to remove no-data
        BSPVNPV = BSPVNPV.where(FC_mask)

        FC_dominant = xarray.Dataset({
            'BS': (BSPVNPV == 0).where(FC_mask),
            'PV': (BSPVNPV == 1).where(FC_mask),
            'NPV': (BSPVNPV == 2).where(FC_mask)
        })

        FC_count = FC_dominant.sum(dim=['x', 'y'])

        # Fractional cover pixel count method
        # Get number of FC pixels, divide by total number of pixels per polygon
        new_ds = xarray.Dataset({
            'BS': (FC_count.BS / total_valid)['BS'] * 100,
            'PV': (FC_count.PV / total_valid)['PV'] * 100,
            'NPV': (FC_count.NPV / total_valid)['NPV'] * 100,
            'Unobservable': (not_pixels / total_valid)['BS'] * 100
        })

        print('dask compute')
        dask_time = default_timer()
        new_ds = new_ds.compute()
        print('dask took', default_timer() - dask_time, 'seconds')
        print(new_ds)

        df = new_ds.to_dataframe()
        df = df.drop('spatial_ref', axis=1)
        df.reset_index(inplace=True)
        return df

    def render_chart(self, df):
        width, height = chart_dimensions(self.style)

        melted = df.melt('time', var_name='Cover Type', value_name='Area')
        melted = melted.dropna()

        style = self.style['table']['columns']

        chart = altair.Chart(melted,
                             width=width,
                             height=height,
                             title='Percentage of Area - Fractional Cover')
        chart = chart.mark_area()
        chart = chart.encode(x='time:T',
                             y=altair.Y('Area:Q', stack='normalize'),
                             color=altair.Color('Cover Type:N',
                                                scale=altair.Scale(domain=self.SHORT_NAMES,
                                                                   range=[style[name]['chartLineColor']
                                                                          for name in self.LONG_NAMES])),
                             tooltip=[altair.Tooltip(field='time', format='%d %B, %Y', title='Date', type='temporal'),
                                      'Area:Q',
                                      'Cover Type:N'])

        return chart

    def render_outputs(self, df, chart):
        return super().render_outputs(df, chart, is_enabled=True, name="FC",
                                      header=self.LONG_NAMES)
